import warnings
warnings.filterwarnings("ignore")

import copy
import torch

from Impeller.model import Impeller
from Impeller.utils import evaluate_Impeller, get_paths

def train(args, data, val_mask, test_mask, x, original_x):

    model = Impeller(in_dim=data.num_node_features, hidden_dim=args.hidden_size, out_dim=data.num_node_features, dropout=args.dropout,
                    num_layers=args.num_layers, num_paths=args.num_paths, path_length=args.path_length, num_edge_types=args.num_edge_types,\
                    alpha=args.alpha, beta=args.beta, operator_type=args.operator_type)
    
    device = torch.device(args.device if torch.cuda.is_available() else 'cpu')
    model = model.to(device)
    data = data.to(device)
    val_mask = val_mask.to(device)
    test_mask = test_mask.to(device)
    x = x.to(device)
    original_x = original_x.to(device)

    print(f'Model structure: {model}, \n\nNumber of parameters: {sum(p.numel() for p in model.parameters() if p.requires_grad)}')

    optimizer = model.setup_optimizer(args.lr, args.weight_decay, args.lr_oc, args.wd_oc)
    paths, path_types = get_paths(args, data)
    data = data.to(device)
        
    criterion = torch.nn.MSELoss()
    best_val_loss = float('inf')
    no_improvement_epochs = 0
    max_no_improvement_epochs = args.patience
    best_model_state = None  # To store the state of the best model

    for epoch in range(args.epochs):
        model.train()
        # paths, path_types = get_paths(args, data)
        optimizer.zero_grad()
        logits = model(x, paths, path_types)
        target_x = x
        # Create a boolean mask of the target where True indicates non-zero values
        mask = target_x != 0
        # Apply the mask to select only the non-zero targets and predictions
        target_x_masked = target_x[mask]
        target_logits_masked = logits[mask]
        # Compute the loss only over the non-zero values
        loss = criterion(target_logits_masked, target_x_masked)
        loss.backward()
        optimizer.step()
        if epoch % args.val_epoch == 0:
            val_l1_distance, val_cosine_sim, val_rmse = evaluate_Impeller(model, x, paths, path_types, device, val_mask, original_x)
        if epoch % args.print_epoch == 0: 
            print(f"Epoch: {epoch}, Loss: {loss.item()}, val_l1_distance: {val_l1_distance}, val_cosine_sim: {val_cosine_sim}, val_rmse: {val_rmse}")
        if val_rmse < best_val_loss:
            best_val_loss = val_rmse
            no_improvement_epochs = 0
            best_model_state = copy.deepcopy(model.state_dict())
        else:
            no_improvement_epochs += 1
        if no_improvement_epochs >= max_no_improvement_epochs:
            print(f"No improvement in validation loss for {max_no_improvement_epochs} epochs, stopping.")
            break
    # Load the best model state before testing
    model.load_state_dict(best_model_state)
    test_l1_distance, test_cosine_sim, test_rmse = evaluate_Impeller(model, x, paths, path_types, device, test_mask, original_x)
    
    return test_l1_distance, test_cosine_sim, test_rmse

def inference(args, data, inference_mask):

    model = Impeller(in_dim=data.num_node_features, hidden_dim=args.hidden_size, out_dim=data.num_node_features, dropout=args.dropout,
                    num_layers=args.num_layers, num_paths=args.num_paths, path_length=args.path_length, num_edge_types=args.num_edge_types,\
                    alpha=args.alpha, beta=args.beta, operator_type=args.operator_type)
    
    device = torch.device(args.device if torch.cuda.is_available() else 'cpu')
    model = model.to(device)
    data = data.to(device)
    inference_mask = inference_mask.to(device)
    x = data.x.to(device)

    print(f'Model structure: {model}, \n\nNumber of parameters: {sum(p.numel() for p in model.parameters() if p.requires_grad)}')

    optimizer = model.setup_optimizer(args.lr, args.weight_decay, args.lr_oc, args.wd_oc)
    paths, path_types = get_paths(args, data)
    data = data.to(device)
    criterion = torch.nn.MSELoss()

    for epoch in range(args.epochs):
        model.train()
        # paths, path_types = get_paths(args, data)
        optimizer.zero_grad()
        logits = model(x, paths, path_types)
        target_x = x
        # Create a boolean mask of the target where True indicates non-zero values
        mask = target_x != 0
        # Apply the mask to select only the non-zero targets and predictions
        target_x_masked = target_x[mask]
        target_logits_masked = logits[mask]
        # Compute the loss only over the non-zero values
        loss = criterion(target_logits_masked, target_x_masked)
        loss.backward()
        optimizer.step()
        if epoch % args.print_epoch == 0: 
            print(f"Epoch: {epoch}, Loss: {loss.item()}")

    return logits